// This is the source file where the methods computing
// the time evolution of Black Holes are implemented.
// Last modification: 22 May 2019
// Authors: Jérémy Auffinger j.auffinger@ipnl.in2p3.fr & Alexandre Arbey alexandre.arbey@ens-lyon.fr

#include "include.h"

double rplus_BH(double M,double a){
	// This function computes the external Kerr radius
	// of a BH of mass M and spin parameter a.
	
	return M*(1. + sqrt(1. - pow(a,2.)));
}

double temp_BH(double M,double a){
	// This function computes the Hawking temperature of a Schwarzschild
	// PBH of mass M and spin parameter a.
	
	return 1./(2.*pi*G)*(rplus_BH(M,a) - M)/(pow(rplus_BH(M,a),2.) + pow(a*M,2.));
}

void read_fM_table(double **fM_table,double *fM_masses,double *fM_a,struct param *parameters){
	// This function reads the f(M,a) factor table used to compute the evolution
	// of BH, stored in the folder /fM_tables.
	
	FILE *table_file;
	char name[64];
	switch(parameters->grav){
		case 0:{ // we use the f(M,a) table computed without the graviton
			sprintf(name,"./src/tables/fM_tables/fM.txt"); // for the moment only with the graviton
			break;
		}
		case 1:{ // we use the f(M,a) table computed taking the graviton into account
			sprintf(name,"./src/tables/fM_tables/fM.txt");
			break;
		}
		default:{
			printf("\n\t [read_fM_table] : ERROR WRONG GRAVITON CHOICE !\n");
			fflush(stdout);
			return;
			break;
		}
	}
	table_file = fopen(name,"r");
	if(!table_file){
		printf("\n\t [read_fM_table] : ERROR COULD NOT READ TABLE %s !\n",name);
		fflush(stdout);
		return;
	}
	rewind(table_file);
	if(parameters->full_output){
		printf("\n\n\t reading:\t %s\n",name);
		fflush(stdout);
	}
	char dummy[20];
	fscanf(table_file,"%s",dummy);
	for(int i = 0;i<parameters->nb_fM_a;i++){
		fscanf(table_file,"%lf",&(fM_a[i]));
	}
	for(int i = 0;i<parameters->nb_fM_masses;i++){
		fscanf(table_file,"%lf",&(fM_masses[i]));
		for(int j = 0;j<parameters->nb_fM_a;j++){
			fscanf(table_file,"%lf",&(fM_table[i][j]));
		}
	}
	fclose(table_file);
	return;
}

void read_gM_table(double **gM_table,double *fM_masses,double *fM_a,struct param *parameters){
	// This function reads the g(M,a) factor table used to compute the evolution
	// of BH, stored in the folder /fM_tables.
	
	FILE *table_file;
	char name[64];
	switch(parameters->grav){
		case 0:{ // we use the g(M,a) table computed without the graviton
			sprintf(name,"./src/tables/fM_tables/gM.txt"); // for the moment only with the graviton
			break;
		}
		case 1:{ // we use the g(M,a) table computed taking the graviton into account
			sprintf(name,"./src/tables/fM_tables/gM.txt");
			break;
		}
		default:{
			printf("\n\t [read_gM_table] : ERROR WRONG GRAVITON CHOICE !\n");
			fflush(stdout);
			return;
			break;
		}
	}
	table_file = fopen(name,"r");
	if(!table_file){
		printf("\n\t [read_gM_table] : ERROR COULD NOT READ TABLE %s !\n",name);
		fflush(stdout);
		return;
	}
	rewind(table_file);
	if(parameters->full_output){
		printf("\t\t\t %s\n\n",name);
		fflush(stdout);
	}
	char dummy[20];
	fscanf(table_file,"%s",dummy);
	for(int i = 0;i<parameters->nb_fM_a;i++){
		fscanf(table_file,"%lf",&(fM_a[i]));
	}
	for(int i = 0;i<parameters->nb_fM_masses;i++){
		fscanf(table_file,"%lf",&(fM_masses[i]));
		for(int j = 0;j<parameters->nb_fM_a;j++){
			fscanf(table_file,"%lf",&(gM_table[i][j]));
		}
	}
	fclose(table_file);
	return;
}

double loss_rate_M(double M,double a,double **fM_table,double *fM_masses,double *fM_a,int counter_M,int counter_a,struct param *parameters){
	// This function computes the BH mass loss rate dM/dt by using the f(M,a) table.
	
	double fM;
	if(parameters->interpolation_method == 0){ // linear interpolation
		fM = fM_table[counter_M][counter_a] + (fM_table[counter_M+1][counter_a] - fM_table[counter_M][counter_a])/(fM_masses[counter_M+1] - fM_masses[counter_M])*(M - fM_masses[counter_M]) + (fM_table[counter_M][counter_a+1] - fM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
	}
	else if(parameters->interpolation_method == 1){ // logarithmic interpolation
		fM = pow(10.,log10(fM_table[counter_M][counter_a]) + (log10(fM_table[counter_M+1][counter_a]) - log10(fM_table[counter_M][counter_a]))/(log10(fM_masses[counter_M+1]) - log10(fM_masses[counter_M]))*(log10(M) - log10(fM_masses[counter_M]))) + (fM_table[counter_M][counter_a+1] - fM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
	}
	else{
		printf("\n\t [loss_rate] : ERROR WRONG INTERPOLATION METHOD !\n");
		fflush(stdout);
		return 1.;
	}
	return -fM/pow(M,2.);
}

double loss_rate_a(double M,double a,double **fM_table,double **gM_table,double *fM_masses,double *fM_a,int counter_M,int counter_a,struct param *parameters){
	// This function computes the BH spin loss rate da/dt by using the f(M,a) and g(M,a) tables.
	
	if(a == 0.){ // no more evolution of the spin
		return 0.;
	}
	else{
		double fM;
		double gM;
		if(parameters->interpolation_method == 0 || counter_a == 0){ // linear interpolation
			fM = fM_table[counter_M][counter_a] + (fM_table[counter_M+1][counter_a] - fM_table[counter_M][counter_a])/(fM_masses[counter_M+1] - fM_masses[counter_M])*(M - fM_masses[counter_M]) + (fM_table[counter_M][counter_a+1] - fM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
			gM = gM_table[counter_M][counter_a] + (gM_table[counter_M+1][counter_a] - gM_table[counter_M][counter_a])/(fM_masses[counter_M+1] - fM_masses[counter_M])*(M - fM_masses[counter_M]) + (gM_table[counter_M][counter_a+1] - gM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
		}
		else if(parameters->interpolation_method == 1){ // logarithmic interpolation
			fM = pow(10.,log10(fM_table[counter_M][counter_a]) + (log10(fM_table[counter_M+1][counter_a]) - log10(fM_table[counter_M][counter_a]))/(log10(fM_masses[counter_M+1]) - log10(fM_masses[counter_M]))*(log10(M) - log10(fM_masses[counter_M]))) + (fM_table[counter_M][counter_a+1] - fM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
			gM = pow(10.,log10(gM_table[counter_M][counter_a]) + (log10(gM_table[counter_M+1][counter_a]) - log10(gM_table[counter_M][counter_a]))/(log10(fM_masses[counter_M+1]) - log10(fM_masses[counter_M]))*(log10(M) - log10(fM_masses[counter_M]))) + (gM_table[counter_M][counter_a+1] - gM_table[counter_M][counter_a])/(fM_a[counter_a+1] - fM_a[counter_a])*(a - fM_a[counter_a]);
		}
		else{
			printf("\n\t [loss_rate] : ERROR WRONG INTERPOLATION METHOD !\n");
			fflush(stdout);
			return 1.;
		}
		return -gM*a/pow(M,3.) + 2.*fM*a/pow(M,3.);
	}
}

void life_evolution(double **life_masses,double **life_spins,double *life_times,double *dts,int *evolution_length,double *init_masses,double *init_spins,double **fM_table,double **gM_table,
	double *fM_masses,double *fM_a,struct param *parameters){
	// This function computes the evolution of BH from their initial mass to the Planck mass.
	// It uses the function loss_rate to compute as estimation of M(t+dt) and a(t+dt) through a Euler
	// method, and adapts the time step to keep relative variations between 0.001 and 0.1.
	// The intermediate masses, spins, times and time intervals are stored in the arrays life_masses[][],
	// life_spins[][], life_times[] and dts[] respectively. A recursion limit is set by the limit parameter
	// but should not be reached.
	
	double dt = parameters->tmin; // current time step
	int current_ind = 0; // current index of computation
	int *counters_M = (int *)malloc(parameters->BHnumber*sizeof(int)); // positions in the f(M,a) / g(M,a) tables
	int *counters_a = (int *)malloc(parameters->BHnumber*sizeof(int)); // positions in the f(M,a) / g(M,a) tables
	double *estimates_M = (double *)malloc(parameters->BHnumber*sizeof(double)); // Euler estimates
	double *estimates_a = (double *)malloc(parameters->BHnumber*sizeof(double)); // Euler estimates
	life_times[0] = parameters->tmin;
	for(int i = 0;i<parameters->BHnumber;i++){
		life_spins[i][0] = init_spins[i];
		life_masses[i][0] = init_masses[i];
		counters_M[i] = parameters->nb_fM_masses-1; // masses are decreasing
		counters_a[i] = parameters->nb_fM_a-1; // spins are decreasing
		evolution_length[i] = 1;
	}
	int counter = 0; // contains the current total number of tried steps
	double evol_min_M;
	double evol_min_a;
	double evol_max_M;
	double evol_max_a;
	int interesting_BH = 0;
	int output_BH = 0;
	if(parameters->full_output){
		printf("\n\n\t computing:");
		fflush(stdout);
	}
	while(counter<parameters->limit-1 && life_masses[parameters->BHnumber-1][current_ind] > 1.1*Mp){ // limit of recursion
	
#if defined(_OPENMP)
#pragma omp parallel for
#endif
		for(int i = 0;i<parameters->BHnumber;i++){
			if(life_masses[i][current_ind] > 1.1*Mp){ // We continue to evolve this mass
				while(counters_M[i] > 0 && life_masses[i][current_ind] < fM_masses[counters_M[i]]){ // we find the closest tabulated value
					counters_M[i]--;
				}
				while(counters_a[i] > 0 && life_spins[i][current_ind] < fM_a[counters_a[i]]){ // we find the closest tabulated value
					counters_a[i]--;
				}
				
				estimates_M[i] = life_masses[i][current_ind] + dt*loss_rate_M(life_masses[i][current_ind],life_spins[i][current_ind],fM_table,fM_masses,fM_a,counters_M[i],counters_a[i],parameters); // Euler estimation of the next mass
				estimates_a[i] = life_spins[i][current_ind] + dt*loss_rate_a(life_masses[i][current_ind],life_spins[i][current_ind],fM_table,gM_table,fM_masses,fM_a,counters_M[i],counters_a[i],parameters); // Euler estimation of the next spin
			}
			else{ // end of the evolution for this initial mass
				if(i+1 < parameters->BHnumber && i+1 >= interesting_BH){
					interesting_BH = i+1;
				}
				estimates_M[i] = 0.;
				estimates_a[i] = 0.;
			}
		}
		evol_min_M = estimates_M[interesting_BH]/life_masses[interesting_BH][current_ind]; // the minimum / maximum variations
		evol_max_M = estimates_M[interesting_BH]/life_masses[interesting_BH][current_ind];
		if(life_spins[interesting_BH][current_ind] != 0.){
			evol_min_a = estimates_a[parameters->BHnumber-1]/life_spins[interesting_BH][current_ind];
			evol_max_a = estimates_a[parameters->BHnumber-1]/life_spins[interesting_BH][current_ind];
		}
		else{
			evol_min_a = 0.;
			evol_max_a = 0.;
		}
		if(evol_max_M < 0.9 || (evol_max_a != 0. && evol_max_a < 0.9)){ // one of the variations is too high, we diminish the time step
			dt = dt/2.;
		}
		else if(evol_min_M > 0.999 && (evol_min_a == 0. || evol_min_a > 0.999) && dt/life_times[current_ind] < 1.){ // all the variations are too low and the time step is not too high compared to the time scale, we increase the time step
			dt = dt*2.;
		}
		else{ // all the variations are fine, we record the evolution and move forward
			current_ind++;
			life_times[current_ind] = life_times[current_ind-1] + dt;
			dts[current_ind] = dt;

			for(int i = 0;i<parameters->BHnumber;i++){
				if(estimates_M[i] > Mp){
					life_masses[i][current_ind] = estimates_M[i];
					if(estimates_a[i] < 1.e-3){ // limit spin under which we consider that it is 0
						life_spins[i][current_ind] = 0.;
					}
					else{
						life_spins[i][current_ind] = estimates_a[i];
					}
					evolution_length[i] = current_ind+1;
				}
				else{
					life_masses[i][current_ind] = 0.;
					life_spins[i][current_ind] = 0.;
				}
			}
		}
		counter++;
		if(parameters->full_output && output_BH != interesting_BH){
			output_BH = interesting_BH;
			if(output_BH%(parameters->BHnumber/10) == 0){
				printf("\t%i/%i",output_BH,parameters->BHnumber);
				fflush(stdout);
			}
		}
	}
	if(parameters->full_output){
		printf("\t%i/%i\n\n",parameters->BHnumber,parameters->BHnumber);
		fflush(stdout);
	}
	if(counters_M!=NULL) free(counters_M);
	if(counters_a!=NULL) free(counters_a);
	if(estimates_M!=NULL) free(estimates_M);
	if(estimates_a!=NULL) free(estimates_a);
	if(counter == parameters->limit-1){
		printf("\n\t [life_evolution] : ERROR ITERATION LIMIT REACHED !\n");
		fflush(stdout);
		return;
	}
	return;
}

void write_life_evolutions(double **life_masses,double **life_spins,double *life_times,int *evolution_length,struct param *parameters){
	// This function writes the number of time steps and the masses and spins as
	// functions of time computed by the function life_evolution for
	// all initial BH masses in the file 'life_evolutions.txt'.
	
	char file_name[500];
	sprintf(file_name,"./results/%s/life_evolutions.txt",parameters->destination_folder);
	FILE *file = fopen(file_name,"w");
	fprintf(file,"Number of timesteps for each BH initial mass.\n\n");
	for(int i = 0;i<parameters->BHnumber;i++){
		fprintf(file,"%10i",evolution_length[i]);
	}
	fprintf(file,"\n\n");
	fprintf(file,"Evolution of the BH masses as functions of time.\n\n");
	fprintf(file,"%15s\n","times/masses");
	for(int i = 0;i<evolution_length[parameters->BHnumber-1];i++){
		fprintf(file,"%15.5e",life_times[i]/time_conversion);
		for(int j = 0;j<parameters->BHnumber;j++){
			fprintf(file,"%15.5e",life_masses[j][i]/mass_conversion);
		}
		fprintf(file,"\n");
	}
	fprintf(file,"\n\n");
	fprintf(file,"Evolution of the BH spins as functions of time.\n\n");
	fprintf(file,"%15s\n","times/spins");
	for(int i = 0;i<evolution_length[parameters->BHnumber-1];i++){
		fprintf(file,"%15.5e",life_times[i]/time_conversion);
		for(int j = 0;j<parameters->BHnumber;j++){
			fprintf(file,"%15.5e",life_spins[j][i]);
		}
		fprintf(file,"\n");
	}
	fclose(file);
	return;
}
